@@ -1,12 +1,9 @@
-/* LICENSE BEGIN
-    This file is part of the SixtyFPS Project -- https://sixtyfps.io
-    Copyright (c) 2021 Olivier Goffart <olivier.goffart@sixtyfps.io>
-    Copyright (c) 2021 Simon Hausmann <simon.hausmann@sixtyfps.io>
+// Copyright © SixtyFPS GmbH <info@sixtyfps.io>
+// SPDX-License-Identifier: (GPL-3.0-only OR LicenseRef-SixtyFPS-commercial)
+
+// Copyright © SixtyFPS GmbH <info@sixtyfps.io>
+// SPDX-License-Identifier: (GPL-3.0-only OR LicenseRef-SixtyFPS-commercial)
 
-    SPDX-License-Identifier: GPL-3.0-only
-    This file is also available under commercial licensing terms.
-    Please contact info@sixtyfps.io for more information.
-LICENSE END */
 #![warn(missing_docs)]
 /*!
     This module contains the event loop implementation using winit, as well as the
@@ -19,7 +16,7 @@ use corelib::layout::Orientation;
 use sixtyfps_corelib as corelib;
 
 use corelib::graphics::Point;
-use corelib::input::{InternalKeyCode, KeyEvent, KeyEventType, KeyboardModifiers, MouseEvent};
+use corelib::input::{KeyEvent, KeyEventType, KeyboardModifiers, MouseEvent};
 use corelib::SharedString;
 use corelib::{window::*, Color};
 use std::cell::{Cell, RefCell};
@@ -62,15 +59,17 @@ pub trait WinitWindow: PlatformWindow {
                 let max_width = constraints_horizontal.max.max(constraints_horizontal.min);
                 let max_height = constraints_vertical.max.max(constraints_vertical.min);
 
+                let sf = self.runtime_window().scale_factor();
+
                 winit_window.set_min_inner_size(if min_width > 0. || min_height > 0. {
-                    Some(winit::dpi::LogicalSize::new(min_width, min_height))
+                    Some(winit::dpi::PhysicalSize::new(min_width * sf, min_height * sf))
                 } else {
                     None
                 });
                 winit_window.set_max_inner_size(if max_width < f32::MAX || max_height < f32::MAX {
-                    Some(winit::dpi::LogicalSize::new(
-                        max_width.min(65535.),
-                        max_height.min(65535.),
+                    Some(winit::dpi::PhysicalSize::new(
+                        (max_width * sf).min(65535.),
+                        (max_height * sf).min(65535.),
                     ))
                 } else {
                     None
@@ -123,7 +122,8 @@ pub trait WinitWindow: PlatformWindow {
             } else {
                 winit_window.set_decorations(true);
             }
-            size = winit_window.inner_size().to_logical(winit_window.scale_factor() as f64);
+            size =
+                winit_window.inner_size().to_logical(self.runtime_window().scale_factor() as f64);
         });
 
         let mut must_resize = false;
@@ -223,13 +223,11 @@ thread_local! {
 
 scoped_tls_hkt::scoped_thread_local!(static CURRENT_WINDOW_TARGET : for<'a> &'a RunningEventLoop<'a>);
 
-#[cfg(not(target_arch = "wasm32"))]
 pub(crate) enum GlobalEventLoopProxyOrEventQueue {
     Proxy(winit::event_loop::EventLoopProxy<CustomEvent>),
     Queue(Vec<CustomEvent>),
 }
 
-#[cfg(not(target_arch = "wasm32"))]
 impl GlobalEventLoopProxyOrEventQueue {
     pub(crate) fn send_event(&mut self, event: CustomEvent) {
         match self {
@@ -253,7 +251,6 @@ impl GlobalEventLoopProxyOrEventQueue {
     }
 }
 
-#[cfg(not(target_arch = "wasm32"))]
 impl Default for GlobalEventLoopProxyOrEventQueue {
     fn default() -> Self {
         Self::Queue(Vec::new())
@@ -265,6 +262,11 @@ pub(crate) static GLOBAL_PROXY: once_cell::sync::OnceCell<
     std::sync::Mutex<GlobalEventLoopProxyOrEventQueue>,
 > = once_cell::sync::OnceCell::new();
 
+#[cfg(target_arch = "wasm32")]
+thread_local! {
+    pub(crate) static GLOBAL_PROXY: RefCell<Option<GlobalEventLoopProxyOrEventQueue>> = RefCell::new(None)
+}
+
 pub(crate) fn with_window_target<T>(callback: impl FnOnce(&dyn EventLoopInterface) -> T) -> T {
     if CURRENT_WINDOW_TARGET.is_set() {
         CURRENT_WINDOW_TARGET.with(|current_target| callback(current_target))
@@ -328,6 +330,22 @@ fn redraw_all_windows() {
     }
 }
 
+mod key_codes {
+    macro_rules! winit_key_to_string_fn {
+        ($($char:literal # $name:ident # $($_qt:ident)|* # $($winit:ident)|* ;)*) => {
+            pub fn winit_key_to_string(virtual_keycode: winit::event::VirtualKeyCode) -> Option<sixtyfps_corelib::SharedString> {
+                let char = match(virtual_keycode) {
+                    $($(winit::event::VirtualKeyCode::$winit => $char,)*)*
+                    _ => return None,
+                };
+                let mut buffer = [0; 6];
+                Some(sixtyfps_corelib::SharedString::from(char.encode_utf8(&mut buffer) as &str))
+            }
+        };
+    }
+    sixtyfps_common::for_each_special_keys!(winit_key_to_string_fn);
+}
+
 fn process_window_event(
     window: Rc<dyn WinitWindow>,
     event: WindowEvent,
@@ -336,6 +354,23 @@ fn process_window_event(
     cursor_pos: &mut Point,
     pressed: &mut bool,
 ) {
+    fn key_event(
+        event_type: KeyEventType,
+        text: SharedString,
+        modifiers: KeyboardModifiers,
+    ) -> KeyEvent {
+        let mut event = KeyEvent { event_type, text, modifiers };
+
+        let tab = String::from(corelib::input::key_codes::Tab);
+
+        // map Shift-Tab into (Shift) Backtab to have a similar behavior as Qt backend
+        if event.text == tab && modifiers.shift {
+            event.text = SharedString::from(String::from(corelib::input::key_codes::Backtab));
+        }
+
+        event
+    }
+
     let runtime_window = window.runtime_window();
     match event {
         WindowEvent::Resized(size) => {
@@ -371,7 +406,8 @@ fn process_window_event(
 
             let modifiers = window.current_keyboard_modifiers().get();
 
-            let mut event = KeyEvent { event_type: KeyEventType::KeyPressed, text, modifiers };
+            let mut event = key_event(KeyEventType::KeyPressed, text, modifiers);
+
             runtime_window.clone().process_key_input(&event);
             event.event_type = KeyEventType::KeyReleased;
             runtime_window.process_key_input(&event);
@@ -385,33 +421,18 @@ fn process_window_event(
         WindowEvent::KeyboardInput { ref input, .. } => {
             corelib::animations::update_animations();
             window.currently_pressed_key_code().set(match input.state {
-                winit::event::ElementState::Pressed => input.virtual_keycode.clone(),
+                winit::event::ElementState::Pressed => input.virtual_keycode,
                 _ => None,
             });
-            if let Some(key_code) =
-                input.virtual_keycode.and_then(|virtual_keycode| match virtual_keycode {
-                    winit::event::VirtualKeyCode::Left => Some(InternalKeyCode::Left),
-                    winit::event::VirtualKeyCode::Right => Some(InternalKeyCode::Right),
-                    winit::event::VirtualKeyCode::Home => Some(InternalKeyCode::Home),
-                    winit::event::VirtualKeyCode::End => Some(InternalKeyCode::End),
-                    winit::event::VirtualKeyCode::Back => Some(InternalKeyCode::Back),
-                    winit::event::VirtualKeyCode::Delete => Some(InternalKeyCode::Delete),
-                    winit::event::VirtualKeyCode::Tab => Some(InternalKeyCode::Tab),
-                    winit::event::VirtualKeyCode::Return
-                    | winit::event::VirtualKeyCode::NumpadEnter => Some(InternalKeyCode::Return),
-                    winit::event::VirtualKeyCode::Escape => Some(InternalKeyCode::Escape),
-                    _ => None,
-                })
-            {
-                let text = key_code.encode_to_string();
-                let event = KeyEvent {
-                    event_type: match input.state {
+            if let Some(text) = input.virtual_keycode.and_then(key_codes::winit_key_to_string) {
+                let event = key_event(
+                    match input.state {
                         winit::event::ElementState::Pressed => KeyEventType::KeyPressed,
                         winit::event::ElementState::Released => KeyEventType::KeyReleased,
                     },
                     text,
-                    modifiers: window.current_keyboard_modifiers().get(),
-                };
+                    window.current_keyboard_modifiers().get(),
+                );
                 runtime_window.process_key_input(&event);
             };
         }
@@ -492,9 +513,11 @@ fn process_window_event(
             runtime_window.process_mouse_input(ev);
         }
         WindowEvent::ScaleFactorChanged { scale_factor, new_inner_size: size } => {
-            let size = size.to_logical(scale_factor);
-            runtime_window.set_window_item_geometry(size.width, size.height);
-            runtime_window.set_scale_factor(scale_factor as f32);
+            if std::env::var("SIXTYFPS_SCALE_FACTOR").is_err() {
+                let size = size.to_logical(scale_factor);
+                runtime_window.set_window_item_geometry(size.width, size.height);
+                runtime_window.set_scale_factor(scale_factor as f32);
+            }
         }
         _ => {}
     }
@@ -513,13 +536,14 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
 
     let event_loop_proxy = not_running_loop_instance.event_loop_proxy;
     #[cfg(not(target_arch = "wasm32"))]
-    {
-        GLOBAL_PROXY
-            .get_or_init(Default::default)
-            .lock()
-            .unwrap()
-            .set_proxy(event_loop_proxy.clone());
-    }
+    GLOBAL_PROXY.get_or_init(Default::default).lock().unwrap().set_proxy(event_loop_proxy.clone());
+    #[cfg(target_arch = "wasm32")]
+    GLOBAL_PROXY.with(|global_proxy| {
+        global_proxy
+            .borrow_mut()
+            .get_or_insert_with(Default::default)
+            .set_proxy(event_loop_proxy.clone())
+    });
 
     let mut winit_loop = not_running_loop_instance.instance;
 
@@ -576,20 +600,19 @@ pub fn run(quit_behavior: sixtyfps_corelib::backend::EventLoopQuitBehavior) {
                 _ => (),
             }
 
-            if *control_flow != winit::event_loop::ControlFlow::Exit {
-                if corelib::animations::CURRENT_ANIMATION_DRIVER
+            if *control_flow != winit::event_loop::ControlFlow::Exit
+                && corelib::animations::CURRENT_ANIMATION_DRIVER
                     .with(|driver| driver.has_active_animations())
-                {
-                    *control_flow = ControlFlow::Poll;
-                    redraw_all_windows()
-                }
+            {
+                *control_flow = ControlFlow::Poll;
+                redraw_all_windows()
             }
 
             corelib::timers::TimerList::maybe_activate_timers();
 
             if *control_flow == winit::event_loop::ControlFlow::Wait {
                 if let Some(next_timer) = corelib::timers::TimerList::next_timeout() {
-                    *control_flow = winit::event_loop::ControlFlow::WaitUntil(next_timer);
+                    *control_flow = winit::event_loop::ControlFlow::WaitUntil(next_timer.into());
                 }
             }
         })
