@@ -1,12 +1,6 @@
-/* LICENSE BEGIN
-    This file is part of the SixtyFPS Project -- https://sixtyfps.io
-    Copyright (c) 2021 Olivier Goffart <olivier.goffart@sixtyfps.io>
-    Copyright (c) 2021 Simon Hausmann <simon.hausmann@sixtyfps.io>
+// Copyright © SixtyFPS GmbH <info@sixtyfps.io>
+// SPDX-License-Identifier: (GPL-3.0-only OR LicenseRef-SixtyFPS-commercial)
 
-    SPDX-License-Identifier: GPL-3.0-only
-    This file is also available under commercial licensing terms.
-    Please contact info@sixtyfps.io for more information.
-LICENSE END */
 // cspell:ignore Noto fontconfig
 
 use femtovg::TextContext;
@@ -301,6 +295,9 @@ impl FontCache {
                 (
                     match source {
                         fontdb::Source::Binary(data) => data.clone(),
+                        // We feed only Source::Binary into fontdb on wasm
+                        #[allow(unreachable_patterns)]
+                        _ => unreachable!(),
                     },
                     face_index,
                 )
@@ -513,7 +510,7 @@ impl FontCache {
         request
             .family
             .as_ref()
-            .map(|family_name| self.available_families.contains(&family_name))
+            .map(|family_name| self.available_families.contains(family_name))
             .unwrap_or(false)
     }
 
