@@ -1,12 +1,6 @@
-/* LICENSE BEGIN
-    This file is part of the SixtyFPS Project -- https://sixtyfps.io
-    Copyright (c) 2021 Olivier Goffart <olivier.goffart@sixtyfps.io>
-    Copyright (c) 2021 Simon Hausmann <simon.hausmann@sixtyfps.io>
+// Copyright © SixtyFPS GmbH <info@sixtyfps.io>
+// SPDX-License-Identifier: (GPL-3.0-only OR LicenseRef-SixtyFPS-commercial)
 
-    SPDX-License-Identifier: GPL-3.0-only
-    This file is also available under commercial licensing terms.
-    Please contact info@sixtyfps.io for more information.
-LICENSE END */
 //! This module contains the GraphicsWindow that used to be within corelib.
 
 // cspell:ignore corelib winit Borderless
@@ -15,13 +9,13 @@ use core::cell::{Cell, RefCell};
 use core::pin::Pin;
 use std::rc::{Rc, Weak};
 
-use super::{ImageCache, ItemGraphicsCache};
+use super::{ItemGraphicsCache, TextureCache};
 use crate::event_loop::WinitWindow;
 use const_field_offset::FieldOffsets;
 use corelib::component::ComponentRc;
 use corelib::graphics::*;
 use corelib::input::KeyboardModifiers;
-use corelib::items::ItemRef;
+use corelib::items::{ItemRef, MouseCursor};
 use corelib::layout::Orientation;
 use corelib::window::{PlatformWindow, PopupWindow, PopupWindowLocation};
 use corelib::Property;
@@ -40,7 +34,9 @@ pub struct GLWindow {
 
     pub(crate) graphics_cache: RefCell<ItemGraphicsCache>,
     // This cache only contains textures. The cache for decoded CPU side images is in crate::IMAGE_CACHE.
-    pub(crate) texture_cache: RefCell<ImageCache>,
+    pub(crate) texture_cache: RefCell<TextureCache>,
+
+    fps_counter: Option<Rc<FPSCounter>>,
 
     #[cfg(target_arch = "wasm32")]
     canvas_id: String,
@@ -64,6 +60,7 @@ impl GLWindow {
             currently_pressed_key_code: Default::default(),
             graphics_cache: Default::default(),
             texture_cache: Default::default(),
+            fps_counter: FPSCounter::new(),
             #[cfg(target_arch = "wasm32")]
             canvas_id,
         })
@@ -142,7 +139,7 @@ impl WinitWindow for GLWindow {
             window.opengl_context.ensure_resized();
 
             {
-                let mut canvas = window.canvas.borrow_mut();
+                let mut canvas = window.canvas.as_ref().unwrap().borrow_mut();
                 // We pass 1.0 as dpi / device pixel ratio as femtovg only uses this factor to scale
                 // text metrics. Since we do the entire translation from logical pixels to physical
                 // pixels on our end, we don't need femtovg to scale a second time.
@@ -157,14 +154,14 @@ impl WinitWindow for GLWindow {
             }
 
             let mut renderer = crate::GLItemRenderer {
-                canvas: window.canvas.clone(),
+                canvas: window.canvas.as_ref().unwrap().clone(),
                 layer_images_to_delete_after_flush: Default::default(),
                 graphics_window: self.clone(),
                 scale_factor,
                 state: vec![crate::State {
                     scissor: Rect::new(
                         Point::default(),
-                        Size::new(size.width as _, size.height as _),
+                        Size::new(size.width as f32, size.height as f32) / scale_factor,
                     ),
                     global_alpha: 1.,
                     layer: None,
@@ -172,11 +169,11 @@ impl WinitWindow for GLWindow {
             };
 
             for (component, origin) in components {
-                corelib::item_rendering::render_component_items(
-                    &component,
-                    &mut renderer,
-                    origin.clone(),
-                );
+                corelib::item_rendering::render_component_items(component, &mut renderer, *origin);
+            }
+
+            if let Some(fps_counter) = &self.fps_counter {
+                fps_counter.measure_frame_rendered(&mut renderer);
             }
 
             renderer.canvas.borrow_mut().flush();
@@ -341,6 +338,11 @@ impl PlatformWindow for GLWindow {
             .with_title(window_title)
             .with_resizable(is_resizable);
 
+        let scale_factor_override = std::env::var("SIXTYFPS_SCALE_FACTOR")
+            .ok()
+            .and_then(|x| x.parse::<f64>().ok())
+            .filter(|f| *f > 0.);
+
         let window_builder = if std::env::var("SIXTYFPS_FULLSCREEN").is_ok() {
             window_builder.with_fullscreen(Some(winit::window::Fullscreen::Borderless(None)))
         } else {
@@ -350,11 +352,16 @@ impl PlatformWindow for GLWindow {
                 layout_info_h.preferred_bounded(),
                 layout_info_v.preferred_bounded(),
             );
+
             if s.width > 0. && s.height > 0. {
                 // Make sure that the window's inner size is in sync with the root window item's
                 // width/height.
                 runtime_window.set_window_item_geometry(s.width, s.height);
-                window_builder.with_inner_size(s)
+                if let Some(f) = scale_factor_override {
+                    window_builder.with_inner_size(s.to_physical::<f32>(f))
+                } else {
+                    window_builder.with_inner_size(s)
+                }
             } else {
                 window_builder
             }
@@ -382,12 +389,50 @@ impl PlatformWindow for GLWindow {
 
         let platform_window = opengl_context.window();
         let runtime_window = self.self_weak.upgrade().unwrap();
-        runtime_window.set_scale_factor(platform_window.scale_factor() as _);
+        runtime_window.set_scale_factor(
+            scale_factor_override.unwrap_or_else(|| platform_window.scale_factor()) as _,
+        );
         let id = platform_window.id();
+
+        if let Some(fps_counter) = &self.fps_counter {
+            cfg_if::cfg_if! {
+                if #[cfg(target_arch = "wasm32")] {
+                    let winsys = "HTML Canvas";
+                } else if #[cfg(any(
+                    target_os = "linux",
+                    target_os = "dragonfly",
+                    target_os = "freebsd",
+                    target_os = "netbsd",
+                    target_os = "openbsd"
+                ))] {
+                    use winit::platform::unix::WindowExtUnix;
+                    let mut winsys = "unknown";
+
+                    #[cfg(feature = "x11")]
+                    if platform_window.xlib_window().is_some() {
+                        winsys = "x11";
+                    }
+
+                    #[cfg(feature = "wayland")]
+                    if platform_window.wayland_surface().is_some() {
+                        winsys = "wayland"
+                    }
+                } else if #[cfg(target_os = "windows")] {
+                    let winsys = "windows";
+                } else if #[cfg(target_os = "macos")] {
+                    let winsys = "macos";
+                } else {
+                    let winsys = "unknown";
+                }
+            }
+
+            fps_counter.start(&format!("GL backend (windowing system: {})", winsys));
+        }
+
         drop(platform_window);
 
         self.map_state.replace(GraphicsWindowBackendState::Mapped(MappedWindow {
-            canvas,
+            canvas: Some(canvas),
             opengl_context,
             clear_color: RgbaColor { red: 255_u8, green: 255, blue: 255, alpha: 255 }.into(),
             constraints: Default::default(),
@@ -400,7 +445,7 @@ impl PlatformWindow for GLWindow {
         // Release GL textures and other GPU bound resources.
         self.with_current_context(|| {
             self.graphics_cache.borrow_mut().clear();
-            self.texture_cache.borrow_mut().remove_textures();
+            self.texture_cache.borrow_mut().clear();
         });
 
         self.map_state.replace(GraphicsWindowBackendState::Unmapped);
@@ -410,6 +455,43 @@ impl PlatformWindow for GLWindow {
         }*/
     }
 
+    fn set_mouse_cursor(&self, cursor: MouseCursor) {
+        let winit_cursor = match cursor {
+            MouseCursor::default => winit::window::CursorIcon::Default,
+            MouseCursor::none => winit::window::CursorIcon::Default,
+            MouseCursor::help => winit::window::CursorIcon::Help,
+            MouseCursor::pointer => winit::window::CursorIcon::Hand,
+            MouseCursor::progress => winit::window::CursorIcon::Progress,
+            MouseCursor::wait => winit::window::CursorIcon::Wait,
+            MouseCursor::crosshair => winit::window::CursorIcon::Crosshair,
+            MouseCursor::text => winit::window::CursorIcon::Text,
+            MouseCursor::alias => winit::window::CursorIcon::Alias,
+            MouseCursor::copy => winit::window::CursorIcon::Copy,
+            MouseCursor::no_drop => winit::window::CursorIcon::NoDrop,
+            MouseCursor::not_allowed => winit::window::CursorIcon::NotAllowed,
+            MouseCursor::grab => winit::window::CursorIcon::Grab,
+            MouseCursor::grabbing => winit::window::CursorIcon::Grabbing,
+            MouseCursor::col_resize => winit::window::CursorIcon::ColResize,
+            MouseCursor::row_resize => winit::window::CursorIcon::RowResize,
+            MouseCursor::n_resize => winit::window::CursorIcon::NResize,
+            MouseCursor::e_resize => winit::window::CursorIcon::EResize,
+            MouseCursor::s_resize => winit::window::CursorIcon::SResize,
+            MouseCursor::w_resize => winit::window::CursorIcon::WResize,
+            MouseCursor::ne_resize => winit::window::CursorIcon::NeResize,
+            MouseCursor::nw_resize => winit::window::CursorIcon::NwResize,
+            MouseCursor::se_resize => winit::window::CursorIcon::SeResize,
+            MouseCursor::sw_resize => winit::window::CursorIcon::SwResize,
+            MouseCursor::ew_resize => winit::window::CursorIcon::EwResize,
+            MouseCursor::ns_resize => winit::window::CursorIcon::NsResize,
+            MouseCursor::nesw_resize => winit::window::CursorIcon::NeswResize,
+            MouseCursor::nwse_resize => winit::window::CursorIcon::NwseResize,
+        };
+        self.with_window_handle(&mut |winit_window| {
+            winit_window.set_cursor_visible(cursor != MouseCursor::none);
+            winit_window.set_cursor_icon(winit_cursor);
+        });
+    }
+
     fn text_size(
         &self,
         font_request: corelib::graphics::FontRequest,
@@ -540,7 +622,7 @@ impl PlatformWindow for GLWindow {
 }
 
 struct MappedWindow {
-    canvas: CanvasRc,
+    canvas: Option<CanvasRc>,
     opengl_context: crate::OpenGLContext,
     clear_color: Color,
     constraints: Cell<(corelib::layout::LayoutInfo, corelib::layout::LayoutInfo)>,
@@ -548,6 +630,15 @@ struct MappedWindow {
 
 impl Drop for MappedWindow {
     fn drop(&mut self) {
+        if let Some(canvas) = self.canvas.take().map(|canvas| Rc::try_unwrap(canvas).ok()) {
+            // The canvas must be destructed with a GL context current, in order to clean up correctly
+            self.opengl_context.with_current_context(|| {
+                drop(canvas);
+            });
+        } else {
+            corelib::debug_log!("internal warning: there are canvas references left when destroying the window. OpenGL resources will be leaked.")
+        }
+
         crate::event_loop::unregister_window(self.opengl_context.window().id());
     }
 }
