@@ -1,12 +1,6 @@
-/* LICENSE BEGIN
-    This file is part of the SixtyFPS Project -- https://sixtyfps.io
-    Copyright (c) 2021 Olivier Goffart <olivier.goffart@sixtyfps.io>
-    Copyright (c) 2021 Simon Hausmann <simon.hausmann@sixtyfps.io>
+// Copyright © SixtyFPS GmbH <info@sixtyfps.io>
+// SPDX-License-Identifier: (GPL-3.0-only OR LicenseRef-SixtyFPS-commercial)
 
-    SPDX-License-Identifier: GPL-3.0-only
-    This file is also available under commercial licensing terms.
-    Please contact info@sixtyfps.io for more information.
-LICENSE END */
 /*!
 
 **NOTE**: This library is an **internal** crate for the [SixtyFPS project](https://sixtyfps.io).
@@ -25,7 +19,9 @@ use std::rc::Rc;
 
 use euclid::approxeq::ApproxEq;
 use event_loop::WinitWindow;
-use sixtyfps_corelib::graphics::{Brush, Color, Image, IntRect, Point, Rect, RenderingCache, Size};
+use sixtyfps_corelib::graphics::{
+    Brush, Color, Image, ImageInner, IntRect, Point, Rect, RenderingCache, Size,
+};
 use sixtyfps_corelib::item_rendering::{CachedRenderingData, ItemRenderer};
 use sixtyfps_corelib::items::{FillRule, ImageFit, ImageRendering};
 use sixtyfps_corelib::properties::Property;
@@ -543,21 +539,15 @@ impl ItemRenderer for GLItemRenderer {
             || {
                 ItemGraphicsCacheEntry::Image({
                     let blur = box_shadow.blur() * self.scale_factor;
-                    let offset_x = box_shadow.offset_x() * self.scale_factor;
-                    let offset_y = box_shadow.offset_y() * self.scale_factor;
                     let width = box_shadow.width() * self.scale_factor;
                     let height = box_shadow.height() * self.scale_factor;
                     let radius = box_shadow.border_radius() * self.scale_factor;
 
-                    let shadow_rect: euclid::Rect<f32, euclid::UnknownUnit> = euclid::rect(
-                        offset_x - blur,
-                        offset_y - blur,
-                        width + 2. * blur,
-                        height + 2. * blur,
-                    );
+                    let shadow_rect: euclid::Rect<f32, euclid::UnknownUnit> =
+                        euclid::rect(0., 0., width + 2. * blur, height + 2. * blur);
 
-                    let shadow_image_width = shadow_rect.max_x().ceil() as usize;
-                    let shadow_image_height = shadow_rect.max_y().ceil() as usize;
+                    let shadow_image_width = shadow_rect.width().ceil() as usize;
+                    let shadow_image_height = shadow_rect.height().ceil() as usize;
 
                     let shadow_image = CachedImage::new_empty_on_gpu(
                         &self.canvas,
@@ -576,13 +566,13 @@ impl ItemRenderer for GLItemRenderer {
                         canvas.clear_rect(
                             0,
                             0,
-                            shadow_rect.max_x().ceil() as u32,
-                            shadow_rect.max_y().ceil() as u32,
+                            shadow_rect.width().ceil() as u32,
+                            shadow_rect.height().ceil() as u32,
                             femtovg::Color::rgba(0, 0, 0, 0),
                         );
 
                         let mut shadow_path = femtovg::Path::new();
-                        shadow_path.rounded_rect(offset_x, offset_y, width, height, radius);
+                        shadow_path.rounded_rect(blur, blur, width, height, radius);
                         canvas.fill_path(
                             &mut shadow_path,
                             femtovg::Paint::color(femtovg::Color::rgb(255, 255, 255)),
@@ -612,12 +602,7 @@ impl ItemRenderer for GLItemRenderer {
                         canvas.global_composite_operation(femtovg::CompositeOperation::SourceIn);
 
                         let mut shadow_image_rect = femtovg::Path::new();
-                        shadow_image_rect.rect(
-                            0.,
-                            0.,
-                            shadow_rect.max_x().ceil(),
-                            shadow_rect.max_y().ceil(),
-                        );
+                        shadow_image_rect.rect(0., 0., shadow_rect.width(), shadow_rect.height());
                         canvas.fill_path(
                             &mut shadow_image_rect,
                             femtovg::Paint::color(to_femtovg_color(&box_shadow.color())),
@@ -649,7 +634,13 @@ impl ItemRenderer for GLItemRenderer {
         let mut shadow_image_rect = femtovg::Path::new();
         shadow_image_rect.rect(0., 0., shadow_image_size.width, shadow_image_size.height);
 
-        self.canvas.borrow_mut().fill_path(&mut shadow_image_rect, shadow_image_paint);
+        self.canvas.borrow_mut().save_with(|canvas| {
+            let blur = box_shadow.blur() * self.scale_factor;
+            let offset_x = box_shadow.offset_x() * self.scale_factor;
+            let offset_y = box_shadow.offset_y() * self.scale_factor;
+            canvas.translate(offset_x - blur, offset_y - blur);
+            canvas.fill_path(&mut shadow_image_rect, shadow_image_paint);
+        });
     }
 
     fn combine_clip(&mut self, mut clip_rect: Rect, mut radius: f32, mut border_width: f32) {
@@ -763,6 +754,19 @@ impl ItemRenderer for GLItemRenderer {
         canvas.fill_path(&mut path, fill_paint);
     }
 
+    fn draw_string(&mut self, string: &str, color: Color) {
+        let font = fonts::FONT_CACHE.with(|cache| {
+            cache.borrow_mut().font(
+                self.graphics_window.default_font_properties(),
+                self.scale_factor,
+                string,
+            )
+        });
+        let paint = font.init_paint(0.0, femtovg::Paint::color(to_femtovg_color(&color)));
+        let mut canvas = self.canvas.borrow_mut();
+        canvas.fill_text(0., 0., string, paint).unwrap();
+    }
+
     fn window(&self) -> WindowRc {
         self.graphics_window.runtime_window()
     }
@@ -905,42 +909,48 @@ impl GLItemRenderer {
             let image_cache_entry =
                 item_cache.get_or_update(&self.graphics_window.graphics_cache, || {
                     let image = source_property.get();
-                    let image_inner = (&image).into();
+                    let image_inner: &ImageInner = (&image).into();
 
-                    ImageCacheKey::new(image_inner, Some(image_rendering))
-                        .and_then(|cache_key| {
-                            self.graphics_window
-                                .texture_cache
-                                .borrow_mut()
-                                .lookup_image_in_cache_or_create(cache_key, || {
-                                    crate::IMAGE_CACHE
-                                        .with(|global_cache| {
-                                            global_cache
-                                                .borrow_mut()
-                                                .load_image_resource(image_inner)
-                                        })
-                                        .and_then(|image| {
-                                            image
-                                                .upload_to_gpu(
-                                                    self, // The condition at the entry of the function ensures that width/height are positive
-                                                    [
-                                                        (target_width.get() * self.scale_factor)
-                                                            as u32,
-                                                        (target_height.get() * self.scale_factor)
-                                                            as u32,
-                                                    ]
-                                                    .into(),
-                                                    image_rendering,
-                                                )
-                                                .map(Rc::new)
-                                        })
-                                })
-                        })
-                        .or_else(|| CachedImage::new_from_resource(image_inner).map(Rc::new))
-                        .map(ItemGraphicsCacheEntry::Image)
-                        .map(|cache_entry| {
-                            self.colorize_image(cache_entry, colorize_property, image_rendering)
-                        })
+                    let target_size_for_scalable_source = image_inner.is_svg().then(|| {
+                        // get the scale factor as a property again, to ensure the cache is invalidated when the scale factor changes
+                        let scale_factor = self.window().scale_factor();
+                        [
+                            (target_width.get() * scale_factor) as u32,
+                            (target_height.get() * scale_factor) as u32,
+                        ]
+                        .into()
+                    });
+
+                    TextureCacheKey::new(
+                        image_inner,
+                        target_size_for_scalable_source,
+                        image_rendering,
+                    )
+                    .and_then(|cache_key| {
+                        self.graphics_window
+                            .texture_cache
+                            .borrow_mut()
+                            .lookup_image_in_cache_or_create(cache_key, || {
+                                crate::IMAGE_CACHE
+                                    .with(|global_cache| {
+                                        global_cache.borrow_mut().load_image_resource(image_inner)
+                                    })
+                                    .and_then(|image| {
+                                        image
+                                            .upload_to_gpu(
+                                                self, // The condition at the entry of the function ensures that width/height are positive
+                                                target_size_for_scalable_source,
+                                                image_rendering,
+                                            )
+                                            .map(Rc::new)
+                                    })
+                            })
+                    })
+                    .or_else(|| CachedImage::new_from_resource(image_inner).map(Rc::new))
+                    .map(ItemGraphicsCacheEntry::Image)
+                    .map(|cache_entry| {
+                        self.colorize_image(cache_entry, colorize_property, image_rendering)
+                    })
                 });
 
             // Check if the image in the cache is loaded. If not, don't draw any image and we'll return
@@ -1225,9 +1235,30 @@ impl sixtyfps_corelib::backend::Backend for Backend {
         #[cfg(not(target_arch = "wasm32"))]
         crate::event_loop::GLOBAL_PROXY.get_or_init(Default::default).lock().unwrap().send_event(e);
         #[cfg(target_arch = "wasm32")]
-        crate::event_loop::with_window_target(|event_loop| {
-            event_loop.event_loop_proxy().send_event(e).ok();
-        })
+        {
+            use wasm_bindgen::closure::Closure;
+            use wasm_bindgen::JsCast;
+
+            let window = web_sys::window().expect("Failed to obtain window");
+            let send_event = || {
+                crate::event_loop::GLOBAL_PROXY.with(|global_proxy| {
+                    global_proxy.borrow_mut().get_or_insert_with(Default::default).send_event(e)
+                });
+            };
+
+            // Calling send_event is usually done by winit at the bottom of the stack,
+            // in event handlers, and thus winit might decide to process the event
+            // immediately within that stack.
+            // To prevent re-entrancy issues that might happen by getting the application
+            // event processed on top of the current stack, mimic the event handling
+            // use case and make sure that our event is processed on top of a clean stack.
+            window
+                .set_timeout_with_callback_and_timeout_and_arguments_0(
+                    &Closure::once_into_js(send_event).as_ref().unchecked_ref(),
+                    0,
+                )
+                .expect("Failed to set timeout");
+        }
     }
 
     fn image_size(&'static self, image: &Image) -> Size {
