@@ -1,12 +1,6 @@
-/* LICENSE BEGIN
-    This file is part of the SixtyFPS Project -- https://sixtyfps.io
-    Copyright (c) 2021 Olivier Goffart <olivier.goffart@sixtyfps.io>
-    Copyright (c) 2021 Simon Hausmann <simon.hausmann@sixtyfps.io>
+// Copyright © SixtyFPS GmbH <info@sixtyfps.io>
+// SPDX-License-Identifier: (GPL-3.0-only OR LicenseRef-SixtyFPS-commercial)
 
-    SPDX-License-Identifier: GPL-3.0-only
-    This file is also available under commercial licensing terms.
-    Please contact info@sixtyfps.io for more information.
-LICENSE END */
 #![cfg(feature = "svg")]
 
 fn with_svg_options<T>(callback: impl FnOnce(usvg::OptionsRef<'_>) -> T) -> T {
@@ -46,7 +40,7 @@ pub fn render(
     let skia_buffer =
         tiny_skia::PixmapMut::from_bytes(buffer.as_mut_slice(), size.width(), size.height())
             .ok_or(usvg::Error::InvalidSize)?;
-    resvg::render(tree, fit, skia_buffer).ok_or(usvg::Error::InvalidSize)?;
+    resvg::render(tree, fit, Default::default(), skia_buffer).ok_or(usvg::Error::InvalidSize)?;
     Ok(image::DynamicImage::ImageRgba8(
         image::RgbaImage::from_raw(size.width(), size.height(), buffer)
             .ok_or(usvg::Error::InvalidSize)?,
